@@ -6,6 +6,13 @@
 
 #include "pllua.h"
 #include "rowstamp.h"
+#include "lua_int64.h"
+#include "rtupdescstk.h"
+
+#include "pllua_pgfunc.h"
+#include "pllua_subxact.h"
+#include "pllua_errors.h"
+
 
 /*
  * [[ Uses of REGISTRY ]]
@@ -26,6 +33,8 @@
 
 /* extended function info */
 typedef struct luaP_Info {
+  RTupDescStack funcxt_wp; /* weak if init_weak used */
+  bool code_storage;
   int oid;
   int vararg;
   Oid result;
@@ -38,7 +47,7 @@ typedef struct luaP_Info {
 /* extended type info */
 typedef struct luaP_Typeinfo {
   int oid;
-  int2 len;
+  int16 len;
   char type;
   char align;
   bool byval;
@@ -85,15 +94,17 @@ static const char PLLUA_DATUM[] = "datum";
 #define resulterror(type) \
   elog(ERROR, "[pllua]: type '%s' (%d) not supported as result", \
       format_type_be(type), (type))
-#define luaP_error(L, tag) \
-        ereport(ERROR, (errcode(ERRCODE_DATA_EXCEPTION), \
-             errmsg("[pllua]: " tag " error"), \
-             errdetail("%s", lua_tostring((L), -1))))
 
 #define datum2string(d, f) \
   DatumGetCString(DirectFunctionCall1((f), (d)))
 #define text2string(d) datum2string((d), textout)
 
+static int luaP_panic(lua_State *L){
+    (void)L;
+    ereport(FATAL, (errmsg("pllua  unhandled exception")));
+    return 0;
+}
+
 /* string2text is simpler, so we implement it here with allocation in upper
  * memory context */
 static Datum string2text (const char *str) {
@@ -115,15 +126,7 @@ static Datum datumcopy (Datum dat, luaP_Typeinfo *ti) {
   return dat;
 }
 
-/* get MemoryContext for state L */
-static MemoryContext luaP_getmemctxt (lua_State *L) {
-  MemoryContext mcxt;
-  lua_pushlightuserdata(L, (void *) L);
-  lua_rawget(L, LUA_REGISTRYINDEX);
-  mcxt = (MemoryContext) lua_touserdata(L, -1);
-  lua_pop(L, 1);
-  return mcxt;
-}
+
 
 
 /* ======= Type ======= */
@@ -136,7 +139,7 @@ static int luaP_typeinfogc (lua_State *L) {
 
 static luaP_Typeinfo *luaP_gettypeinfo (lua_State *L, int oid) {
   luaP_Typeinfo *ti;
-  lua_pushinteger(L, oid);
+  lua_push_oidstring(L, oid);
   lua_rawget(L, LUA_REGISTRYINDEX);
   if (lua_isnil(L, -1)) { /* not cached? */
     HeapTuple type;
@@ -157,7 +160,7 @@ static luaP_Typeinfo *luaP_gettypeinfo (lua_State *L, int oid) {
     fmgr_info_cxt(typeinfo->typinput, &ti->input, mcxt);
     fmgr_info_cxt(typeinfo->typoutput, &ti->output, mcxt);
     ti->tupdesc = NULL;
-    if (ti->type == 'c') { /* complex? */
+    if (ti->type == TYPTYPE_COMPOSITE) {
       TupleDesc td = lookup_rowtype_tupdesc(oid, typeinfo->typtypmod);
       MemoryContext m = MemoryContextSwitchTo(mcxt);
       ti->tupdesc = CreateTupleDescCopyConstr(td);
@@ -169,7 +172,7 @@ static luaP_Typeinfo *luaP_gettypeinfo (lua_State *L, int oid) {
     lua_pushlightuserdata(L, (void *) PLLUA_TYPEINFO);
     lua_rawget(L, LUA_REGISTRYINDEX); /* Typeinfo_MT */
     lua_setmetatable(L, -2);
-    lua_pushinteger(L, oid);
+    lua_push_oidstring(L, oid);
     lua_pushvalue(L, -2);
     lua_rawset(L, LUA_REGISTRYINDEX); /* REG[oid] = typeinfo */
     lua_pop(L, 2); /* nil and typeinfo */
@@ -233,6 +236,8 @@ static luaP_Datum *luaP_pushrawdatum (lua_State *L, Datum dat,
 
 static void luaP_preptrigger (lua_State *L, TriggerData *tdata) {
   const char *relname;
+  char *namespace;
+
   lua_pushglobaltable(L);
   lua_pushstring(L, PLLUA_TRIGGERVAR);
   lua_newtable(L);
@@ -266,35 +271,39 @@ static void luaP_preptrigger (lua_State *L, TriggerData *tdata) {
   else
     elog(ERROR, "[pllua]: unknown trigger 'operation' event");
   lua_setfield(L, -2, "operation");
-  /* relation (name) */
+
+  /* relation */
   relname = NameStr(tdata->tg_relation->rd_rel->relname);
-  lua_getfield(L, LUA_REGISTRYINDEX, relname);
-  if (lua_isnil(L, -1)) { /* not cached? */
-    lua_pop(L, 1);
-    lua_createtable(L, 0, 2);
-    lua_pushstring(L, relname);
-    lua_setfield(L, -2, "name");
-    luaP_pushdesctable(L, tdata->tg_relation->rd_att);
-    lua_pushinteger(L, (int) tdata->tg_relation->rd_id);
-    lua_pushvalue(L, -2); /* attribute table */
-    lua_rawset(L, LUA_REGISTRYINDEX); /* cache desc */
-    lua_setfield(L, -2, "attributes");
-    lua_pushvalue(L, -1);
-    lua_setfield(L, LUA_REGISTRYINDEX, relname); /* cache relation */
-  }
+  namespace = get_namespace_name(tdata->tg_relation->rd_rel->relnamespace);
+
+  lua_createtable(L, 0, 3);
+  lua_pushstring(L, relname);
+  lua_setfield(L, -2, "name");
+  luaP_pushdesctable(L, tdata->tg_relation->rd_att);
+  lua_push_oidstring(L, (int) tdata->tg_relation->rd_id);
+  lua_pushvalue(L, -2); /* attribute table */
+
+  lua_rawset(L, LUA_REGISTRYINDEX); /* cache desc */
+  lua_setfield(L, -2, "attributes");
+  lua_pushinteger(L, (int) tdata->tg_relation->rd_id);
+  lua_setfield(L, -2, "oid");
+  lua_pushstring(L, namespace);
+  lua_setfield(L, -2, "namespace");
+
   lua_setfield(L, -2, "relation");
+
   /* row */
   if (TRIGGER_FIRED_FOR_ROW(tdata->tg_event)) {
     if (TRIGGER_FIRED_BY_UPDATE(tdata->tg_event)) {
-      luaP_pushtuple(L, tdata->tg_relation->rd_att, tdata->tg_newtuple,
+      luaP_pushtuple_trg(L, tdata->tg_relation->rd_att, tdata->tg_newtuple,
           tdata->tg_relation->rd_id, 0);
       lua_setfield(L, -2, "row"); /* new row */
-      luaP_pushtuple(L, tdata->tg_relation->rd_att, tdata->tg_trigtuple,
+      luaP_pushtuple_trg(L, tdata->tg_relation->rd_att, tdata->tg_trigtuple,
           tdata->tg_relation->rd_id, 1);
       lua_setfield(L, -2, "old"); /* old row */
     }
     else { /* insert or delete */
-      luaP_pushtuple(L, tdata->tg_relation->rd_att, tdata->tg_trigtuple,
+      luaP_pushtuple_trg(L, tdata->tg_relation->rd_att, tdata->tg_trigtuple,
           tdata->tg_relation->rd_id, 0);
       lua_setfield(L, -2, "row"); /* old row */
     }
@@ -317,6 +326,7 @@ static Datum luaP_gettriggerresult (lua_State *L) {
 }
 
 static void luaP_cleantrigger (lua_State *L) {
+  rtds_tryclean(rtds_get_current()); //fi->functx;
   lua_pushglobaltable(L);
   lua_pushstring(L, PLLUA_TRIGGERVAR);
   lua_pushnil(L);
@@ -343,7 +353,7 @@ static int luaP_modinit (lua_State *L) {
     else {
       status = 0;
       if (SPI_processed > 0) { /* any rows? */
-        int i;
+        unsigned int i;
         for (i = 0; i < SPI_processed; i++) {
           bool isnull;
           /* push module name */
@@ -413,29 +423,33 @@ static int luaP_print (lua_State *L) {
   return 0;
 }
 
+
+#define PLLUA_REPORT(elevel) do{\
+    if (lua_type(L, 1)== LUA_TTABLE){\
+        luatable_report(L, elevel);\
+        return 0;\
+    }\
+    luaL_checkstring(L, 1);\
+    ereport(elevel, (errmsg("%s", lua_tostring(L, 1))));\
+    return 0;\
+    }while(0)
+
 static int luaP_info (lua_State *L) {
-  luaL_checkstring(L, 1);
-  info(lua_tostring(L, 1));
-  return 0;
+    PLLUA_REPORT(INFO);
 }
 
 static int luaP_log (lua_State *L) {
-  luaL_checkstring(L, 1);
-  ereport(LOG, (errmsg("%s", lua_tostring(L, 1))));
-  return 0;
+    PLLUA_REPORT(LOG);
 }
 
 static int luaP_notice (lua_State *L) {
-  luaL_checkstring(L, 1);
-  ereport(NOTICE, (errmsg("%s", lua_tostring(L, 1))));
-  return 0;
+    PLLUA_REPORT(NOTICE);
 }
 
 static int luaP_warning (lua_State *L) {
-  luaL_checkstring(L, 1);
-  ereport(WARNING, (errmsg("%s", lua_tostring(L, 1))));
-  return 0;
+    PLLUA_REPORT(WARNING);
 }
+#undef PLLUA_REPORT
 
 static int luaP_fromstring (lua_State *L) {
   int oid = luaP_gettypeoid(luaL_checkstring(L, 1));
@@ -444,21 +458,40 @@ static int luaP_fromstring (lua_State *L) {
   int inoid = oid;
   Datum v;
   /* from getTypeIOParam in lsyscache.c */
-  if (ti->type == 'b' && OidIsValid(ti->elem)) inoid = ti->elem;
+  if (ti->type == TYPTYPE_BASE && OidIsValid(ti->elem)) inoid = ti->elem;
   v = InputFunctionCall(&ti->input, (char *) s, inoid, 0); /* typmod = 0 */
   luaP_pushdatum(L, v, oid);
   return 1;
 }
 
+#ifdef PLLUA_DEBUG
+static int
+luaP_memstat(lua_State *L)
+{
+	(void)L;
+	MemoryContextStats(TopMemoryContext);
+	return 0;
+}
+#endif
+
 static const luaL_Reg luaP_funcs[] = {
-  {"setshared", luaP_setshared},
-  {"log", luaP_log},
-  {"print", luaP_print},
-  {"info", luaP_info},
-  {"notice", luaP_notice},
-  {"warning", luaP_warning},
-  {"fromstring", luaP_fromstring},
-  {NULL, NULL}
+    {"assert", luaB_assert},
+    {"error", luaB_error},
+    {"fromstring", luaP_fromstring},
+    {"info", luaP_info},
+    {"log", luaP_log},
+#ifdef PLLUA_DEBUG
+    {"memstat", luaP_memstat},
+#endif
+    {"notice", luaP_notice},
+    {"pcall", subt_luaB_pcall},
+    {"pgfunc", get_pgfunc},
+    {"print", luaP_print},
+    {"setshared", luaP_setshared},
+    {"subtransaction", use_subtransaction},
+    {"warning", luaP_warning},
+    {"xpcall", subt_luaB_xpcall},
+    {NULL, NULL}
 };
 
 void luaP_close (lua_State *L) {
@@ -469,17 +502,25 @@ void luaP_close (lua_State *L) {
 
 lua_State *luaP_newstate (int trusted) {
   int status;
+
+
   MemoryContext mcxt = AllocSetContextCreate(TopMemoryContext,
       "PL/Lua context", ALLOCSET_DEFAULT_MINSIZE, ALLOCSET_DEFAULT_INITSIZE,
       ALLOCSET_DEFAULT_MAXSIZE);
   lua_State *L = luaL_newstate();
+  lua_atpanic(L, luaP_panic);
   /* version */
   lua_pushliteral(L, PLLUA_VERSION);
   lua_setglobal(L, "_PLVERSION");
   /* memory context */
-  lua_pushlightuserdata(L, (void *) L);
+  lua_pushlightuserdata(L, p_lua_mem_cxt);
   lua_pushlightuserdata(L, (void *) mcxt);
   lua_rawset(L, LUA_REGISTRYINDEX);
+
+  lua_pushlightuserdata(L, p_lua_master_state);
+  lua_pushlightuserdata(L, (void *) L);
+  lua_rawset(L, LUA_REGISTRYINDEX);
+
   /* core libs */
   if (trusted) {
     const luaL_Reg luaP_trusted_libs[] = {
@@ -488,11 +529,16 @@ lua_State *luaP_newstate (int trusted) {
 #else
       {"_G", luaopen_base},
       {LUA_COLIBNAME, luaopen_coroutine},
+#if LUA_VERSION_NUM < 503
       {LUA_BITLIBNAME, luaopen_bit32},
+#endif
 #endif
       {LUA_TABLIBNAME, luaopen_table},
       {LUA_STRLIBNAME, luaopen_string},
       {LUA_MATHLIBNAME, luaopen_math},
+#ifdef LUA_JITLIBNAME
+      {LUA_JITLIBNAME, luaopen_jit},
+#endif
       {LUA_OSLIBNAME, luaopen_os}, /* restricted */
       {LUA_LOADLIBNAME, luaopen_package}, /* just for pllua.init modules */
       {NULL, NULL}
@@ -500,6 +546,7 @@ lua_State *luaP_newstate (int trusted) {
     const char *os_funcs[] = {"date", "clock", "time", "difftime", NULL};
     const luaL_Reg *reg = luaP_trusted_libs;
     const char **s = os_funcs;
+
     for (; reg->func; reg++) {
 #if LUA_VERSION_NUM <= 501
       lua_pushcfunction(L, reg->func);
@@ -522,6 +569,10 @@ lua_State *luaP_newstate (int trusted) {
   }
   else
     luaL_openlibs(L);
+
+  register_error_mt(L);
+  register_funcinfo_mt(L);
+  register_int64(L);
   /* setup typeinfo and raw datum MTs */
   lua_pushlightuserdata(L, (void *) PLLUA_TYPEINFO);
   lua_newtable(L); /* luaP_Typeinfo MT */
@@ -550,14 +601,16 @@ lua_State *luaP_newstate (int trusted) {
   lua_pushglobaltable(L);
   luaP_register(L, luaP_funcs);
   lua_pop(L, 1);
+
   /* SPI */
   luaP_registerspi(L);
   lua_setglobal(L, PLLUA_SPIVAR);
   if (trusted) {
+	
     const char *package_keys[] = { /* to be removed */
       "preload", "loadlib", "loaders", "seeall", NULL};
     const char *global_keys[] = { /* to be removed */
-      "require", "module", "dofile", "loadfile", NULL};
+      "require", "module", "dofile", "loadfile", "jit", NULL};
     const char **s;
     /* clean package module */
     lua_getglobal(L, "package");
@@ -595,25 +648,36 @@ static luaP_Info *luaP_newinfo (lua_State *L, int nargs, int oid,
   luaP_Info *fi;
   int i;
   luaP_Typeinfo *ti;
+  bool code_storage = ((nargs == 1)
+                       &&(argtype[0] == INTERNALOID)
+                       &&(rettype == INTERNALOID));
+
+
   fi = lua_newuserdata(L, sizeof(luaP_Info) + nargs * sizeof(Oid));
+  fi->funcxt_wp = NULL;
   fi->oid = oid;
-  /* read arg types */
-  for (i = 0; i < nargs; i++) {
-    ti = luaP_gettypeinfo(L, argtype[i]);
-    if (ti->type == 'p') /* pseudo-type? */
-      ereport(ERROR,
-          (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-           errmsg("[pllua]: functions cannot take type '%s'",
-          format_type_be(argtype[i]))));
-    fi->arg[i] = argtype[i];
+  fi->code_storage = code_storage;
+  if(!code_storage){
+      /* read arg types */
+      for (i = 0; i < nargs; i++) {
+        ti = luaP_gettypeinfo(L, argtype[i]);
+        if (ti->type == TYPTYPE_PSEUDO)
+          ereport(ERROR,
+              (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+               errmsg("[pllua]: functions cannot take type '%s'",
+              format_type_be(argtype[i]))));
+        fi->arg[i] = argtype[i];
+      }
+      /* read result type */
+      ti = luaP_gettypeinfo(L, rettype);
+      if (ti->type == TYPTYPE_PSEUDO && rettype != VOIDOID && rettype != TRIGGEROID)
+        ereport(ERROR,
+            (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+             errmsg("[pllua]: functions cannot return type '%s'",
+               format_type_be(rettype))));
+  }else{
+      fi->arg[0] = INTERNALOID;
   }
-  /* read result type */
-  ti = luaP_gettypeinfo(L, rettype);
-  if (ti->type == 'p' && rettype != VOIDOID && rettype != TRIGGEROID) 
-    ereport(ERROR,
-        (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-         errmsg("[pllua]: functions cannot return type '%s'",
-           format_type_be(rettype))));
   fi->vararg = rettype == TRIGGEROID; /* triggers are vararg */
   fi->result = rettype;
   fi->result_isset = isset;
@@ -629,10 +693,11 @@ static void luaP_newfunction (lua_State *L, int oid, HeapTuple proc,
   Form_pg_proc procst;
   bool isnull;
   Datum prosrc, *argname;
-  const char *s, *fname;
+  const char *source, *fname;
   text *t;
   luaL_Buffer b;
   int init = (*fi == NULL); /* not initialized? */
+  const char *chunk_name = NULL;//PLLUA_CHUNKNAME;
   /* read proc info */
   procst = (Form_pg_proc) GETSTRUCT(proc);
   prosrc = SysCacheGetAttr(PROCOID, proc, Anum_pg_proc_prosrc, &isnull);
@@ -640,12 +705,12 @@ static void luaP_newfunction (lua_State *L, int oid, HeapTuple proc,
   nargs = procst->pronargs;
   /* get info userdata */
   if (init) {
-    lua_pushinteger(L, oid);
+    lua_push_oidstring(L, oid);
     *fi = luaP_newinfo(L, nargs, oid, procst);
   }
   lua_pushlightuserdata(L, (void *) *fi);
   /* check #argnames */
-  if (nargs > 0) {
+  if ((nargs > 0)&&((*fi)->code_storage == 0)) {
     int nnames;
     Datum argnames = SysCacheGetAttr(PROCOID, proc,
         Anum_pg_proc_proargnames, &isnull);
@@ -674,14 +739,16 @@ static void luaP_newfunction (lua_State *L, int oid, HeapTuple proc,
   luaL_addlstring(&b, fname, strlen(fname));
   luaL_addlstring(&b, "=function(", 10);
   /* read arg names */
-  if ((*fi)->vararg) luaL_addlstring(&b, "...", 3);
-  else {
-    int i;
-    for (i = 0; i < nargs; i++) {
-      if (i > 0) luaL_addchar(&b, ',');
-      t = DatumGetTextP(argname[i]);
-      luaL_addlstring(&b, VARDATA(t), VARSIZE(t) - VARHDRSZ);
-    }
+  if((*fi)->code_storage == 0){
+      if ((*fi)->vararg) luaL_addlstring(&b, "...", 3);
+      else {
+        int i;
+        for (i = 0; i < nargs; i++) {
+          if (i > 0) luaL_addchar(&b, ',');
+          t = DatumGetTextP(argname[i]);
+          luaL_addlstring(&b, VARDATA(t), VARSIZE(t) - VARHDRSZ);
+        }
+      }
   }
   luaL_addlstring(&b, ") ", 2);
   /* read source */
@@ -692,11 +759,20 @@ static void luaP_newfunction (lua_State *L, int oid, HeapTuple proc,
   luaL_addlstring(&b, fname, strlen(fname));
   /* create function */
   luaL_pushresult(&b);
-  s = lua_tostring(L, -1);
-  if (luaL_loadbuffer(L, s, strlen(s), PLLUA_CHUNKNAME))
-    luaP_error(L, "compile");
+  source = lua_tostring(L, -1);
+
+
+#if defined(PLLUA_DEBUG)
+      chunk_name = source;
+#else
+      chunk_name = fname;
+#endif
+
+
+  if (luaL_loadbuffer(L, source, strlen(source), chunk_name))
+    luapg_error(L, "compile");
   lua_remove(L, -2); /* source */
-  if (lua_pcall(L, 0, 1, 0)) luaP_error(L, "call");
+  if (lua_pcall(L, 0, 1, 0)) luapg_error(L, "call");
   rowstamp_set(&(*fi)->stamp, proc); /* row-stamp info */
   lua_pushvalue(L, -1); /* func */
   if (init) {
@@ -717,7 +793,7 @@ static luaP_Info *luaP_pushfunction (lua_State *L, int oid) {
   proc = SearchSysCache(PROCOID, ObjectIdGetDatum((Oid) oid), 0, 0, 0);
   if (!HeapTupleIsValid(proc))
     elog(ERROR, "[pllua]: cache lookup failed for function %u", (Oid) oid);
-  lua_pushinteger(L, oid);
+  lua_push_oidstring(L, oid);
   lua_rawget(L, LUA_REGISTRYINDEX);
   if (lua_isnil(L, -1)) { /* not interned? */
     lua_pop(L, 1); /* nil */
@@ -791,6 +867,9 @@ void luaP_pushdatum (lua_State *L, Datum dat, Oid type) {
     case INT4OID:
       lua_pushinteger(L, (lua_Integer) DatumGetInt32(dat));
       break;
+    case INT8OID:
+        setInt64lua(L,(DatumGetInt64(dat)));
+        break;
     case TEXTOID:
       lua_pushstring(L, text2string(dat));
       break;
@@ -806,10 +885,14 @@ void luaP_pushdatum (lua_State *L, Datum dat, Oid type) {
       else lua_pushnil(L);
       break;
     }
+    case RECORDOID:
+      luaP_pushrecord(L, dat);
+      break;
     default: {
-      luaP_Typeinfo *ti = luaP_gettypeinfo(L, type);
+      luaP_Typeinfo *ti;
+      ti = luaP_gettypeinfo(L, type);
       switch (ti->type) {
-        case 'c': { /* complex? */
+        case TYPTYPE_COMPOSITE: {
           HeapTupleHeader tup = DatumGetHeapTupleHeader(dat);
           int i;
           const char *key;
@@ -827,11 +910,11 @@ void luaP_pushdatum (lua_State *L, Datum dat, Oid type) {
           }
           break;
         }
-        case 'p': /* pseudo? */
+        case TYPTYPE_PSEUDO:
           if (type != VOIDOID) argerror(type);
           break;
-        case 'b': /* base? */
-        case 'd': /* domain? */
+        case TYPTYPE_BASE:
+        case TYPTYPE_DOMAIN:
           if (ti->elem != 0 && ti->len == -1) { /* array? */
             ArrayType *arr = DatumGetArrayTypeP(dat);
             char *p = ARR_DATA_PTR(arr);
@@ -845,7 +928,7 @@ void luaP_pushdatum (lua_State *L, Datum dat, Oid type) {
             luaP_pushrawdatum(L, dat, ti);
           break;
 #if PG_VERSION_NUM >= 80300
-        case 'e': /* enum? */
+        case TYPTYPE_ENUM:
           lua_pushinteger(L, (lua_Integer) DatumGetInt32(dat)); /* 4-byte long */
           break;
 #endif
@@ -916,7 +999,7 @@ static int luaP_getarraydims (lua_State *L, int *ndims, int *dims,
       }
       else {
         bool isnull;
-        Datum d = luaP_todatum(L, typeelem, typmod, &isnull);
+        Datum d = luaP_todatum(L, typeelem, typmod, &isnull, -1);
         Pointer v = DatumGetPointer(d);
         n = 0;
         if (ti->len == -1) /* varlena? */
@@ -946,7 +1029,7 @@ static void luaP_toarray (lua_State *L, char **p, int ndims,
     for (i = 0; i < (*dims); i++) {
       Pointer v;
       lua_rawgeti(L, -1, (*lb) + i);
-      v = DatumGetPointer(luaP_todatum(L, typeelem, typmod, &isnull));
+      v = DatumGetPointer(luaP_todatum(L, typeelem, typmod, &isnull, -1));
       if (!isnull) {
         *bitval |= *bitmask;
         if (ti->len > 0) {
@@ -990,57 +1073,61 @@ static void luaP_toarray (lua_State *L, char **p, int ndims,
   }
 }
 
-Datum luaP_todatum (lua_State *L, Oid type, int typmod, bool *isnull) {
+Datum luaP_todatum (lua_State *L, Oid type, int typmod, bool *isnull, int idx) {
   Datum dat = 0; /* NULL */
-  *isnull = lua_isnil(L, -1);
+  *isnull = lua_isnil(L, idx);
   if (!(*isnull || type == VOIDOID)) {
     switch (type) {
       /* base and domain types */
       case BOOLOID:
-        dat = BoolGetDatum(lua_toboolean(L, -1));
+        dat = BoolGetDatum(lua_toboolean(L, idx));
         break;
       case FLOAT4OID:
-        dat = Float4GetDatum((float4) lua_tonumber(L, -1));
+        dat = Float4GetDatum((float4) lua_tonumber(L, idx));
         break;
       case FLOAT8OID:
-        dat = Float8GetDatum((float8) lua_tonumber(L, -1));
+        dat = Float8GetDatum((float8) lua_tonumber(L, idx));
         break;
       case INT2OID:
-        dat = Int16GetDatum(lua_tointeger(L, -1));
+        dat = Int16GetDatum(lua_tointeger(L, idx));
         break;
       case INT4OID:
-        dat = Int32GetDatum(lua_tointeger(L, -1));
+        dat = Int32GetDatum(lua_tointeger(L, idx));
+        break;
+      case INT8OID:
+        dat = Int64GetDatum(get64lua(L, idx));
         break;
       case TEXTOID: {
-        const char *s = lua_tostring(L, -1);
+        const char *s = lua_tostring(L, idx);
         if (s == NULL) elog(ERROR,
             "[pllua]: string expected for datum conversion, got %s",
-            lua_typename(L, lua_type(L, -1)));
+            lua_typename(L, lua_type(L, idx)));
         dat = string2text(s);
         break;
       }
       case REFCURSOROID: {
-        Portal cursor = luaP_tocursor(L, -1);
+        Portal cursor = luaP_tocursor(L, idx);
         dat = string2text(cursor->name);
         break;
       }
       default: {
-        luaP_Typeinfo *ti = luaP_gettypeinfo(L, type);
+        luaP_Typeinfo *ti;
+        ti = luaP_gettypeinfo(L, type);
         switch (ti->type) {
-          case 'c': /* complex? */
-            if (lua_type(L, -1) == LUA_TTABLE) {
+          case TYPTYPE_COMPOSITE:
+            if (lua_type(L, idx) == LUA_TTABLE) {
               int i;
               luaP_Buffer *b;
-              if (lua_type(L, -1) != LUA_TTABLE)
+              if (lua_type(L, idx) != LUA_TTABLE)
                 elog(ERROR, "[pllua]: table expected for record result, got %s",
-                    lua_typename(L, lua_type(L, -1)));
+                    lua_typename(L, lua_type(L, idx)));
               /* create tuple */
               b = luaP_getbuffer(L, ti->tupdesc->natts);
               for (i = 0; i < ti->tupdesc->natts; i++) {
-                lua_getfield(L, -1, NameStr(ti->tupdesc->attrs[i]->attname));
+                lua_getfield(L, idx, NameStr(ti->tupdesc->attrs[i]->attname));
                 /* only simple types allowed in record */
                 b->value[i] = luaP_todatum(L, ti->tupdesc->attrs[i]->atttypid,
-                    ti->tupdesc->attrs[i]->atttypmod, b->null + i);
+                    ti->tupdesc->attrs[i]->atttypmod, b->null + i, idx);
                 lua_pop(L, 1);
               }
               /* make copy in upper executor memory context */
@@ -1052,24 +1139,24 @@ Datum luaP_todatum (lua_State *L, Oid type, int typmod, bool *isnull) {
               if (tuple == NULL)
                 elog(ERROR,
                     "[pllua]: table or tuple expected for record result, got %s",
-                    lua_typename(L, lua_type(L, -1)));
+                    lua_typename(L, lua_type(L, idx)));
               dat = PointerGetDatum(SPI_returntuple(tuple, ti->tupdesc));
             }
             break;
-          case 'b': /* base? */
-          case 'd': /* domain? */
-            if (ti->elem != 0 && ti->len == -1) { /* array? */
+          case TYPTYPE_BASE:
+          case TYPTYPE_DOMAIN:
+            if (ti->elem != 0 && ti->len == idx) { /* array? */
               luaP_Typeinfo *te;
               int ndims, dims[MAXDIM], lb[MAXDIM];
               int i, size;
               bool hasnulls;
               ArrayType *a;
-              if (lua_type(L, -1) != LUA_TTABLE)
+              if (lua_type(L, idx) != LUA_TTABLE)
                 elog(ERROR,
                     "[pllua]: table expected for array conversion, got %s",
-                    lua_typename(L, lua_type(L, -1)));
+                    lua_typename(L, lua_type(L, idx)));
               te = luaP_gettypeinfo(L, ti->elem);
-              for (i = 0; i < MAXDIM; i++) dims[i] = lb[i] = -1;
+              for (i = 0; i < MAXDIM; i++) dims[i] = lb[i] = idx;
               size = luaP_getarraydims(L, &ndims, dims, lb, te, ti->elem,
                   typmod, &hasnulls);
               if (size == 0) { /* empty array? */
@@ -1115,19 +1202,19 @@ Datum luaP_todatum (lua_State *L, Oid type, int typmod, bool *isnull) {
               dat = PointerGetDatum(a);
             }
             else {
-              luaP_Datum *d = luaP_toudata(L, -1, PLLUA_DATUM);
+              luaP_Datum *d = luaP_toudata(L, idx, PLLUA_DATUM);
               if (d == NULL) elog(ERROR,
                   "[pllua]: raw datum expected for datum conversion, got %s",
-                  lua_typename(L, lua_type(L, -1)));
+                  lua_typename(L, lua_type(L, idx)));
               dat = datumcopy(d->datum, ti);
             }
             break;
 #if PG_VERSION_NUM >= 80300
-          case 'e': /* enum? */
-            dat = Int32GetDatum(lua_tointeger(L, -1));
+          case TYPTYPE_ENUM:
+            dat = Int32GetDatum(lua_tointeger(L, idx));
             break;
 #endif
-          case 'p': /* pseudo? */
+          case TYPTYPE_PSEUDO:
           default:
             resulterror(type);
         }
@@ -1139,7 +1226,7 @@ Datum luaP_todatum (lua_State *L, Oid type, int typmod, bool *isnull) {
 
 static Datum luaP_getresult (lua_State *L, FunctionCallInfo fcinfo,
     Oid type) {
-  Datum dat = luaP_todatum(L, type, 0, &fcinfo->isnull);
+  Datum dat = luaP_todatum(L, type, 0, &fcinfo->isnull, -1);
   lua_settop(L, 0);
   return dat;
 }
@@ -1147,11 +1234,12 @@ static Datum luaP_getresult (lua_State *L, FunctionCallInfo fcinfo,
 
 /* ======= luaP_callhandler ======= */
 
-static void luaP_cleanthread (lua_State *L, lua_State **thread) {
-  lua_pushlightuserdata(L, (void *) *thread);
-  lua_pushnil(L);
-  lua_rawset(L, LUA_REGISTRYINDEX);
-  *thread = NULL;
+static void luaP_cleanthread (lua_State *L, lua_State **thread, luaP_Info *fi) {
+    fi->funcxt_wp = rtds_free_if_not_used(fi->funcxt_wp);
+    lua_pushlightuserdata(L, (void *) *thread);
+    lua_pushnil(L);
+    lua_rawset(L, LUA_REGISTRYINDEX);
+    *thread = NULL;
 }
 
 Datum luaP_validator (lua_State *L, Oid oid) {
@@ -1178,19 +1266,31 @@ Datum luaP_validator (lua_State *L, Oid oid) {
 
 Datum luaP_callhandler (lua_State *L, FunctionCallInfo fcinfo) {
   Datum retval = 0;
+  int base = 0;
   luaP_Info *fi;
+  RTupDescStack prev;
   bool istrigger;
   if (SPI_connect() != SPI_OK_CONNECT)
     elog(ERROR, "[pllua]: could not connect to SPI manager");
   istrigger = CALLED_AS_TRIGGER(fcinfo);
   fi = luaP_pushfunction(L, (int) fcinfo->flinfo->fn_oid);
+  if (fi->code_storage == 1){
+    luaL_error(L, "attempt to call non-callable function");
+  }
+
+  if (fi->funcxt_wp == NULL){
+    fi->funcxt_wp = rtds_initStack_weak(L, &fi->funcxt_wp);
+  }
+  rtds_inuse(fi->funcxt_wp);
+
+  prev = rtds_set_current(fi->funcxt_wp);
   PG_TRY();
   {
     if ((fi->result == TRIGGEROID && !istrigger)
         || (fi->result != TRIGGEROID && istrigger))
       ereport(ERROR,
-          (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-           errmsg("[pllua]: trigger function can only be called as trigger")));
+              (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+               errmsg("[pllua]: trigger function can only be called as trigger")));
     if (istrigger) {
       TriggerData *trigdata = (TriggerData *) fcinfo->context;
       int i, nargs;
@@ -1198,7 +1298,14 @@ Datum luaP_callhandler (lua_State *L, FunctionCallInfo fcinfo) {
       nargs = trigdata->tg_trigger->tgnargs;
       for (i = 0; i < nargs; i++) /* push args */
         lua_pushstring(L, trigdata->tg_trigger->tgargs[i]);
-      if (lua_pcall(L, nargs, 0, 0)) luaP_error(L, "runtime");
+      //trigger call
+      if (lua_pcall(L, nargs, 0, 0)) {
+#if defined(PLLUA_DEBUG)
+        luapg_error(L, getLINE());
+#else
+        luapg_error(L, "runtime");
+#endif
+      }
       if (TRIGGER_FIRED_FOR_ROW(trigdata->tg_event)
           && TRIGGER_FIRED_BEFORE(trigdata->tg_event)) /* return? */
         retval = luaP_gettriggerresult(L);
@@ -1212,9 +1319,9 @@ Datum luaP_callhandler (lua_State *L, FunctionCallInfo fcinfo) {
           if (!rsi || !IsA(rsi, ReturnSetInfo)
               || (rsi->allowedModes & SFRM_ValuePerCall) == 0)
             ereport(ERROR,
-                (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-                 errmsg("[pllua]: set-valued function called in context"
-                   "that cannot accept a set")));
+                    (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+                     errmsg("[pllua]: set-valued function called in context"
+                            "that cannot accept a set")));
           rsi->returnMode = SFRM_ValuePerCall;
           fi->L = lua_newthread(L);
           lua_pushlightuserdata(L, (void *) fi->L);
@@ -1224,12 +1331,14 @@ Datum luaP_callhandler (lua_State *L, FunctionCallInfo fcinfo) {
         }
         lua_xmove(L, fi->L, 1); /* function */
         luaP_pushargs(fi->L, fcinfo, fi);
+
 #if LUA_VERSION_NUM <= 501
         status = lua_resume(fi->L, fcinfo->nargs);
 #else
         status = lua_resume(fi->L, fi->L, fcinfo->nargs);
 #endif
-        hasresult = !lua_isnoneornil(fi->L, 1);
+        rtds_notinuse(fi->funcxt_wp);
+        hasresult = !lua_isnone(fi->L, 1);
         if (status == LUA_YIELD && hasresult) {
           rsi->isDone = ExprMultipleResult; /* SRF: next */
           retval = luaP_getresult(fi->L, fcinfo, fi->result);
@@ -1238,14 +1347,37 @@ Datum luaP_callhandler (lua_State *L, FunctionCallInfo fcinfo) {
           rsi->isDone = ExprEndResult; /* SRF: done */
           fcinfo->isnull = true;
           retval = (Datum) 0;
-          luaP_cleanthread(L, &fi->L);
+          luaP_cleanthread(L, &fi->L, fi);
+        }
+        else {
+#if defined(PLLUA_DEBUG)
+          luapg_error(fi->L, getLINE());
+#else
+          luapg_error(fi->L, "runtime");
+#endif
         }
-        else luaP_error(fi->L, "runtime");
       }
       else {
+        int status = 0;
+
         luaP_pushargs(L, fcinfo, fi);
-        if (lua_pcall(L, fcinfo->nargs, 1, 0))
-          luaP_error(L, "runtime");
+        base = lua_gettop(L) - fcinfo->nargs;  /* function index */
+        lua_pushcfunction(L, traceback);  /* push traceback function */
+        lua_insert(L, base);  /* put it under chunk and args */
+        //func call
+        status = lua_pcall(L, fcinfo->nargs, 1, base);
+        lua_remove(L, base);  /* remove traceback function */
+        if (status){
+
+          fi->funcxt_wp = rtds_unref(fi->funcxt_wp);
+#if defined(PLLUA_DEBUG)
+          luapg_error(L, getLINE());
+#else
+          luapg_error(L, "runtime");
+#endif
+        }
+        fi->funcxt_wp = rtds_unref(fi->funcxt_wp);
+
         retval = luaP_getresult(L, fcinfo, fi->result);
       }
     }
@@ -1254,9 +1386,10 @@ Datum luaP_callhandler (lua_State *L, FunctionCallInfo fcinfo) {
   PG_CATCH();
   {
     if (L != NULL) {
-      luaP_cleantrigger(L);
+      luaP_cleantrigger(L);//fi->funcxt ref--
+
       if (fi->result_isset && fi->L != NULL) /* clean thread? */
-        luaP_cleanthread(L, &fi->L);
+        luaP_cleanthread(L, &fi->L, fi);
       lua_settop(L, 0); /* clear Lua stack */
     }
     fcinfo->isnull = true;
@@ -1264,29 +1397,76 @@ Datum luaP_callhandler (lua_State *L, FunctionCallInfo fcinfo) {
     PG_RE_THROW();
   }
   PG_END_TRY();
+  rtds_set_current(prev);
   if (SPI_finish() != SPI_OK_FINISH)
     elog(ERROR, "[pllua]: could not disconnect from SPI manager");
   return retval;
 }
 
+
+
+
+#include "rtupdesc.h"
 #if PG_VERSION_NUM >= 90000
 /* ======= luaP_inlinehandler ======= */
 
 Datum luaP_inlinehandler (lua_State *L, const char *source) {
+  RTupDescStack funcxt;
+  RTupDescStack prev;
+  int base = 0;
+  int status = 0;
   if (SPI_connect() != SPI_OK_CONNECT)
     elog(ERROR, "[pllua]: could not connect to SPI manager");
+
+  funcxt = rtds_initStack(L);
+  rtds_inuse(funcxt);
+
+  prev = rtds_set_current(funcxt);
+
   PG_TRY();
   {
-    if (luaL_loadbuffer(L, source, strlen(source), PLLUA_CHUNKNAME))
-      luaP_error(L, "compile");
-    if (lua_pcall(L, 0, 0, 0)) luaP_error(L, "runtime");
+    const char *chunk_name = "anonymous";//PLLUA_CHUNKNAME;
+
+#if defined(PLLUA_DEBUG)
+    chunk_name = source;
+#endif
+
+    if (luaL_loadbuffer(L, source, strlen(source), chunk_name))
+      luapg_error(L, "compile");
+
+    base = lua_gettop(L) ;  /* function index */
+    lua_pushcfunction(L, traceback);  /* push traceback function */
+    lua_insert(L, base);  /* put it under chunk and args */
+    status = lua_pcall(L, 0, 0, base);
+    lua_remove(L, base);  /* remove traceback function */
   }
   PG_CATCH();
   {
-    if (L != NULL) lua_settop(L, 0); /* clear Lua stack */
+    funcxt = rtds_unref(funcxt);
+    rtds_set_current(prev);
+
+    if (L != NULL) {
+      lua_settop(L, 0); /* clear Lua stack */
+      lua_gc(L, LUA_GCCOLLECT, 0);
+    }
     PG_RE_THROW();
   }
   PG_END_TRY();
+
+  funcxt = rtds_unref(funcxt);
+  rtds_set_current(prev);
+
+  if (status) {
+    lua_gc(L, LUA_GCCOLLECT, 0);
+#if defined(PLLUA_DEBUG)
+    setLINE(AT);
+    luapg_error(L, getLINE());
+#else
+    luapg_error(L, "runtime");
+#endif
+  }
+
+
   if (SPI_finish() != SPI_OK_FINISH)
     elog(ERROR, "[pllua]: could not disconnect from SPI manager");
   PG_RETURN_VOID();
